class SmartCounter(object):
    def __init__(self, data=None):
        self.db = {}

        data = data or {}
        for k, v in data.items():
            self.inc(k, v)

    def inc(self, mode, count=1):
        self.db[mode] = self.db.get(mode, 0) + count

    def add(self, obj_2):
        for k, v in obj_2.db.items():
            self.db[k] = self.db.get(k, 0) + v

    def __str__(self):
        return ", ".join([f"{k}: {v}" for k, v in self.db.items()])



import os

from ROTools.Helpers.Attr import setattr_ex, getattr_ex
from ROTools.Helpers.DictObj import DictObj

class Config(DictObj):
    def __init__(self, **kwargs):
        super().__init__()
        for k in kwargs.keys():
            setattr(self, k, kwargs[k])

    def add_env_data(self, prefix):
        extra_config = [(key[len(prefix):].lower(), value) for key, value in os.environ.items() if key.startswith(prefix)]

        for key, value in extra_config:
            _old_value = getattr_ex(self, key)
            if _old_value is not None and isinstance(_old_value, bool):
                setattr_ex(self, key, value in ["true", "True", "TRUE", "1", ])
                continue

            if _old_value is not None and isinstance(_old_value, int):
                setattr_ex(self, key, int(value))
                continue

            if _old_value is not None and isinstance(_old_value, float):
                setattr_ex(self, key, float(value))
                continue

            setattr_ex(self, key, value)

    def dump_config(self):
        print()
        print("========")
        self.dump()
        print("========")
        print()



import os
import sys
from datetime import datetime, timezone

import humanize


def print_info():
    python_ver = sys.version.replace('\n', ' ')
    print()
    print(f"Build version    \t: {os.getenv('RR_BUILD_VERSION', 'undefined')}")
    print(f"Build time UTC   \t: {os.getenv('RR_BUILD_TIME', 'undefined')}")
    print(f"Current time UTC \t: {datetime.now(timezone.utc)}")
    if os.getenv('RR_BUILD_TIME') is not None:
        build_old = humanize.naturaltime(datetime.now(timezone.utc) - datetime.fromisoformat(os.getenv('RR_BUILD_TIME')))
        print(f"Build old\t\t: {build_old}")
    print(f"PWD              \t: {os.getcwd()}")
    print(f"Python           \t: {python_ver}")
    print("---")
    print()



import sys
import threading
import time
import traceback

import requests

_local_sessions = threading.local()


def get_session_data():
    if not hasattr(_local_sessions, "session"):
        _local_sessions.session = (requests.Session(), 0)

    session, counter = _local_sessions.session
    counter += 1
    _local_sessions.session = (session, counter)

    return session, counter


def make_request_wrapper(request_cb=None, sleep_times=(1, 5, 10, 15, 20, 25, 30, 30, None), **all_params):
    for sleep_time in sleep_times:
        try:
            return request_cb(**all_params)
        except Exception as e:
            print(f"Error [{repr(e)}] and sleep for {sleep_time}", file=sys.stderr)
            traceback.print_exc()
            if sleep_time is None:
                raise e
            time.sleep(sleep_time)
    raise Exception("Unknown error")



import os
import signal
import sys
import time
from multiprocessing import Process, Event


class WorkersCollection:
    def __init__(self):
        self.stop_event = Event()
        self.process = []
        self._is_running = False

    def add(self, worker_class, count, start=False):

        for index in range(count):
            process = Process(target=worker_class.start, args=(index, self.stop_event,))
            if start:
                process.start()
            self.process.append(process)

    def stop(self, ):
        self.stop_event.set()

    def run(self, monitor_cb, monitor_refresh_time):
        def stop_app():
            print(f"Parent process (PID: {os.getpid()}) received SIGTERM. Waiting for workers to finish...")
            self.stop()

        signal.signal(signal.SIGTERM, lambda a, b: stop_app())
        signal.signal(signal.SIGINT, lambda a, b: stop_app())

        self.start()
        self.monitor(cb=monitor_cb, monitor_refresh_time=monitor_refresh_time)
        self.join()

    def start(self, ):
        for process in self.process:
            if process.exitcode is None and not process.is_alive():
                process.start()

    def join(self, ):
        self.stop()
        for item in self.process:
            item.join()

    def is_running(self):
        return all([a.is_alive() for a in self.process])

    def check_error(self):
        error = len([a for a in self.process if a.exitcode is not None and a.exitcode != 0]) > 0
        if error:
            self.stop()

    def monitor(self, cb, monitor_refresh_time):
        try:
            while self.is_running():
                time.sleep(monitor_refresh_time)
                cb()
        except KeyboardInterrupt:
            print("\nMain process received KeyboardInterrupt, terminating workers...")

            self.stop()
            self.join()
            print("All workers terminated. Main process exiting.")
        except Exception as e:
            print("Monitor Error", file=sys.stderr)
            self.stop()
            raise e



class RedisSingletonLock:
    def __init__(self, redis, lock_timeout=30):
        self.redis = redis
        self.lock_key = 'app_lock'
        self.lock_timeout = lock_timeout
        self.lock_acquired = False

    def acquire_lock(self):
        is_locked = self.redis.set(self.lock_key, "LOCKED", nx=True, ex=self.lock_timeout)
        if is_locked:
            self.lock_acquired = True
            return
        raise RuntimeError("Another instance is already running.")

    def refresh_lock(self):
        if self.lock_acquired:
            self.redis.expire(self.lock_key, self.lock_timeout)

    def release_lock(self):
        if self.lock_acquired:
            self.redis.delete(self.lock_key)
            self.lock_acquired = False



import signal
import threading


class SignalHandler:
    _instance = None

    def __init__(self, use_exception=False):
        if self._instance is not None:
            raise Exception("This class is a singleton!")

        SignalHandler._instance = self

        self.stop_event = threading.Event()

        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    @classmethod
    def get(cls):
        if cls._instance is None:
            cls()
        return cls._instance

    def exit_gracefully(self, *args):
        print("SIGNAL STOP")
        self.stop_event.set()

    def is_exit_app(self):
        return self.stop_event.is_set()



import time


class SleepController:
    def __init__(self, single_sleep, wake_up_cb):
        self._single_sleep = single_sleep
        self._wake_up_cb = wake_up_cb
        if self._wake_up_cb is None:
            self._wake_up_cb = lambda: False

    def sleep_for(self, sleep_time):
        _start_time = time.time()
        _delta = 0
        while _delta < sleep_time and not self._wake_up_cb():
            time.sleep(self._single_sleep)
            _delta = time.time() - _start_time



import copy

from ROTools.Helpers.Attr import setattr_ex, hasattr_ex
from ROTools.Helpers.DumpBase import DumpBase


def _gen_fields_list(obj, current_path=None):
    if not isinstance(obj, DictObj):
        return [(current_path, obj)]
    result = []
    for key, value in obj.items():
        next_path = ".".join([a for a in (current_path, key) if a is not None])

        if isinstance(value, DictObj):
            result.extend(_gen_fields_list(value, current_path=next_path))
        else:
            result.append((next_path, value))
    return result


class DictObj(DumpBase):
    def __init__(self, d=None):
        if d is None:
            return
        if isinstance(d, DictObj):
            for k, v in d.__dict__.items():
                self.__dict__[k] = copy.deepcopy(v)
            return

        if isinstance(d, dict):
            self._build_dict(d)
            return

        raise Exception("Flow")

    def _build_dict(self, d):
        for a, b in d.items():
            if isinstance(b, (list, tuple)):
                setattr(self, a, [DictObj(x) if isinstance(x, dict) else x for x in b])
                continue
            setattr(self, a, DictObj(b) if isinstance(b, dict) else b)

    def get(self, name, default=None):
        return getattr(self, name, default)

    def set(self, path, value):
        return setattr_ex(self, path, value, parent_class=DictObj)

    def to_dict(self):
        result = {}
        for k, v in self.__dict__.items():
            if isinstance(v, DictObj):
                v = v.to_dict()
            result[k] = v
        return result

    def set_values(self, values, only_if_exists=True):
        for path, nev_value in values:
            if only_if_exists and not hasattr_ex(self, path):
                continue
            self.set(path, nev_value)

    def fields_list(self):
        return _gen_fields_list(self)

    def keys(self):
        return self.__dict__.keys()

    def items(self):
        return self.__dict__.items()

    def clone(self):
        import copy
        return copy.deepcopy(self)



import time


class StopWatch:
    def __init__(self):
        self._start_time = time.time()
        self._last_stop_time = self._start_time
        self._named_periods = {}
        self.saved = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.stop()

    def __repr__(self):
        return f"{self.saved:>2.1f}"

    def store(self, name):
        _time = time.time()
        self._named_periods[name] = _time - self._last_stop_time
        self._last_stop_time = _time

    def get_time(self, name):
        return self._named_periods[name]

    def get_until_now(self):
        return time.time() - self._start_time

    def is_more_than_restart(self, time_limit_sec):
        _delta = time.time() - self._start_time
        if _delta < time_limit_sec:
            return False
        self._start_time = time.time()
        return True

    def stop(self):
        self.saved = time.time() - self._start_time



import os

from ROTools.Config.Config import Config
from ROTools.Helpers.DictObj import DictObj


def _config_constructor(loader, node):
    fields = loader.construct_mapping(node, deep=True)
    for item in fields.items():
        if isinstance(item[1], dict):
            fields[item[0]] = DictObj(item[1])
    return Config(**fields)


class ConfigBuilder:
    def __init__(self, config_directory, config_file, env_prefix):
        entries = os.listdir(config_directory)
        files = [file for file in entries if os.path.isfile(os.path.join(config_directory, file))]
        files = [os.path.join(config_directory, a) for a in files if a.endswith(('.yaml', '.yml'))]

        main_file_name = os.path.join(config_directory, config_file)
        if main_file_name not in files:
            raise Exception("Config file not found!")

        files = [a for a in files if a != main_file_name]

        self.config = self.load_yaml(main_file_name, tag='!Config')
        for file in files:
            sub_config = self.load_yaml(file, tag='!Config')
            for key, value in sub_config.items():
                self.config.set(key, value)

        if env_prefix is not None:
            self.config.add_env_data(prefix=env_prefix)

    def load_yaml(self, file_name, tag=None):
        import yaml
        yaml.add_constructor(tag, _config_constructor)

        with open(file_name, 'r') as file:
            content = file.read()
        config = yaml.load(content, Loader=yaml.FullLoader)
        return config



import io
import json

from minio.error import S3Error


class MinioWrapper:
    def __init__(self, minio_client, bucket_name=None):
        self.client = minio_client
        self.bucket_name = bucket_name

    def get_json(self, object_name):
        from minio import S3Error
        try:
            response = self.client.get_object(self.bucket_name, object_name)
            object_data = response.read().decode('utf-8')
            response.close()
            response.release_conn()
            return json.loads(object_data)
        except S3Error as err:
            if err.code == 'NoSuchKey':
                return None
            raise err

    def object_exists(self, object_name):
        try:
            self.client.stat_object(self.bucket_name, object_name)
            return True
        except S3Error as err:
            if err.code == 'NoSuchKey':
                return False
            raise

    def list_objects(self, prefix=None, recursive=False):
        return self.client.list_objects(self.bucket_name, prefix=prefix, recursive=recursive)

    def get_metadata(self, object_name):
        stat = self.client.stat_object(self.bucket_name, object_name)
        return stat.metadata

    def put_json(self, object_name, json_obj, metadata=None):
        data_bytes = io.BytesIO(json.dumps(json_obj, ensure_ascii=False, indent=2).encode("utf-8"))
        _len = data_bytes.getbuffer().nbytes
        self.client.put_object(self.bucket_name, object_name, data_bytes, length=_len, content_type="application/json", metadata=metadata)
        obj = self.client.stat_object(self.bucket_name, object_name)
        if obj.object_name != object_name:
            raise Exception("put data error")

    def put_object(self, object_name, data, content_type, metadata=None):
        data_bytes = io.BytesIO(data)
        _len = data_bytes.getbuffer().nbytes

        self.client.put_object(self.bucket_name, object_name, data_bytes, length=_len, content_type=content_type, metadata=metadata)
        obj = self.client.stat_object(self.bucket_name, object_name)
        if obj.object_name != object_name:
            raise Exception("put data error")

    def delete_objects(self, prefix=None, recursive=False):
        to_delete = list(self.list_objects(prefix=prefix, recursive=recursive))
        to_delete = [obj.object_name for obj in to_delete]
        for item in to_delete:
            self.delete_object(object_name=item)

    def delete_object(self, object_name=None):
        return self.client.remove_objects(self.bucket_name, object_name)

    def create_bucket(self, bucket_name):
        bucket_name = bucket_name or self.bucket_name
        if not self.client.bucket_exists(bucket_name):
            self.client.make_bucket(bucket_name)



class DumpBase(object):

    def dump(self, indent_level=0):
        sp = "  " * indent_level

        for k, value in self.__dict__.items():
            if isinstance(value, DumpBase):
                print(f"{sp}{k}:")
                value.dump(indent_level + 1)
                continue

            if isinstance(value, list) and all([isinstance(a, DumpBase) for a in value]):
                print(f"{sp}{k}:")
                for sub_value in value:
                    sub_value.dump(indent_level + 1)  # todo fix it. list display
                continue

            print(f"{sp}{k:<22}: {value}")



from setuptools import setup, find_packages

setup(
    name="rotools",
    version='0.1.67',
    author="Robert Olechowski",
    author_email="robertolechowski@gmail.com",
    description="Robert Olechowski python tools",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/RobertOlechowski/ROTools",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    install_requires=[],
)


import time
from datetime import datetime


class RateLimiter:
    def __init__(self, request_delay, show_wait=True):
        self.request_delay = request_delay
        self.last_call = datetime.fromisoformat("1971-01-01")
        self.show_wait = show_wait

    def call_wait(self):
        _delta = (datetime.now() - self.last_call).total_seconds()
        _wait_time = (self.request_delay / 1000) - _delta
        if _wait_time <= 0:
            self.last_call = datetime.now()
            return
        if self.show_wait:
            print(f"Wait for {_wait_time:>1.2f} seconds")
        time.sleep(_wait_time)
        self.last_call = datetime.now()



def getattr_ex(_obj, name):
    if isinstance(name, str):
        name = name.split(".")

    if isinstance(name, list):
        for item in name:
            if _obj is None:
                return None
            _obj = getattr(_obj, item, None)
        return _obj

    raise Exception("Param error")


def setattr_ex(_obj, name, value, parent_class=object):
    if isinstance(name, str):
        name = name.split(".")

    if isinstance(name, list):
        for item in name[:-1]:
            _next = getattr(_obj, item, None)
            if _next is None:
                _next = parent_class()
                setattr(_obj, item, _next)
            _obj = _next
        setattr(_obj, name[-1], value)
        return

    raise Exception("Param error")


def hasattr_ex(_obj, name):
    if isinstance(name, str):
        name = name.split(".")

    if isinstance(name, list):
        for i, item in enumerate(name):
            is_last = i == len(name) - 1
            if hasattr(_obj, item) is False:
                return False
            _obj = getattr(_obj, item, None)
            if _obj is None and not is_last:
                return False
        return True

    raise Exception("Param error")
